######################################################################
##        Copyright (c) 2020 Carsten Wulff Software, Norway
## ###################################################################
## Created       : wulff at 2020-10-24
## ###################################################################
##  The MIT License (MIT)
##
##  Permission is hereby granted, free of charge, to any person obtaining a copy
##  of this software and associated documentation files (the "Software"), to deal
##  in the Software without restriction, including without limitation the rights
##  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
##  copies of the Software, and to permit persons to whom the Software is
##  furnished to do so, subject to the following conditions:
##
##  The above copyright notice and this permission notice shall be included in all
##  copies or substantial portions of the Software.
##
##  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
##  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
##  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
##  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
##  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
##  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
##  SOFTWARE.
##
######################################################################

import os
import re


class colors:
    '''Colors class:
    reset all colors with colors.reset
    
    '''
    reset='\033[0m'
    bold='\033[01m'
    disable='\033[02m'
    fg = {
        "red" : '\033[31m',
        "green" : '\033[32m',
        "blue" : '\033[34m',
        "cyan" :'\033[36m',
        "yellow" :'\033[93m',
        "default": '',
    }

class Command:

    def __init__(self,verbose=False):
        self.indentstr = "|-"
        self.indent = 0
        self.verbose = verbose

    def getColor(self,color):
        if(color in colors.fg):
            return colors.fg[color]
        else:
            return colors.fg["default"]

    def getIndent(self,nextindent=None):

        if(nextindent is None):
            nextindent = self.indent
        
        if(nextindent == 0):
            return ""
        else:
            return self.indentstr + self.getIndent(nextindent-1)

    def comment(self,ss,color="green"):
        print(self.getIndent() + self.getColor(color) + ss  + colors.reset)

    def warning(self,ss):
        self.comment(ss,"yellow")

    def error(self,ss):
        ss_h = "Error: "
        self.comment(ss_h + ss,"red")

    def doCmd(self,cmd):
        os.system(cmd)

    def sub(self,buffer,keyval):
        """Replace ${NAME} constructs in buffer with values from 'dict' or shell environment"""
        for (k,v) in keyval.items():
            buffer = re.sub('\${%s}'%k,v,buffer)
        m = re.search('\${([^}]+)}',buffer)
        if(m):
            for var in m.groups():
                val = os.getenv(var)
                if(val):
                    buffer = re.sub('\${%s}' %var,val,buffer)
        return buffer



#!/usr/bin/env python3

import click
import cicconf

@click.group()
@click.option("--config",default="config.yaml",help="Configuration file")
@click.option("--verbose/--no-verbose",default=False,help="Turn on extra output")
@click.pass_context
def cli(ctx,config,verbose):
    ctx.ensure_object(dict)
    c = cicconf.Config(config,verbose)
    ctx.obj["c"] = c
    pass

@cli.command()
@click.option("--https",is_flag=True,default=False,help="Use https for clone (override for git@)")
@click.option("--onclone/--no-onclone",is_flag=True,default=True,help="Don't run on_clone event")
@click.pass_context
def clone(ctx,https,onclone):
    """Clone repositories in config file"""
    c = ctx.obj["c"]

    if(c.read()):
        c.clone(useHttps=https,doOnClone=onclone)
        pass

@cli.command()
@click.argument("name")
@click.pass_context
def newip(ctx,name):
    """Create a new IP with name <project>_<name>_<technology>
    The <project> and <technology> is fetched from the config file.

    The name should be short and must be unique (for example, bias, aic01 or similar).
    """

    c = ctx.obj["c"]
    if(c.read()):
        c.newIp(name)


@cli.command()
@click.pass_context
def status(ctx):
    """
    Report the status of each of the configured IPs
    """
    c = ctx.obj["c"]
    if(c.read()):
        c.status()

@cli.command()
@click.pass_context
def update(ctx):
    """
    Update all ips to correct branch according to config file
    """
    c = ctx.obj["c"]
    if(c.read()):
        c.update()

if __name__ == "__main__":
    cli(obj={})



#!/usr/bin/env python3

import git
import cicconf
import os
import yaml

class Repo(cicconf.Command):

    def __init__(self,name,obj,verbose=False):
        super().__init__(verbose)
        self.name = name
        self.remote = None
        self.revision = None
        self.on_clone = None


        if("revision" in obj):
            self.revision = obj["revision"]

        if("remote" in obj):
            self.remote = obj["remote"]

        if("on_clone" in obj):
            self.on_clone = obj["on_clone"]

    def exists(self):
        if(os.path.exists(self.name)):
            return True
        return False

    def clone(self,useHttp=False,doOnClone=True ):
        if(self.exists()):
            self.comment(f"{self.name} already exists")
            return

        if(not self.remote):
            self.error(f"No remote found for {self.name}")
            return

        src = self.remote
        if(useHttp and "git@" in src):
            src = src.replace(":","/")
            src = src.replace("git@","https://")


        self.comment(f"%-25s: Cloning {src}" % (self.name))
        r = git.Repo.clone_from(src, self.name)

        if(self.revision):
            self.comment(f"%-25s: Checkout {self.revision}" % (self.name))
            r.git.checkout(self.revision)

        if(self.on_clone and doOnClone):
            self.doCmd(f"cd {self.name};{self.on_clone}")

    def status(self):

        if(not self.exists()):
            self.warning(f"%-25s%-15s%-15s" % (self.name,"","uncloned"))
            return


        repo = git.Repo(self.name)

        status = ""

        #- Check branch
        if(repo.head.is_detached):
            branch = ""
            sha = repo.head.commit.hexsha
            for t in repo.tags:
                if(t.object.hexsha == sha):
                    branch = t.name

            if(branch == ""):
                branch = "unknown"
        else:
            try:
                branch = repo.active_branch
            except Exception as e:
                self.error(" Could not find branch name of {self.name}")
                return


        if(str(branch) != self.revision):
            status += f"{branch} "


        #- Check ahead/behind
        if(not repo.head.is_detached):
            commits_behind = repo.iter_commits(f"{branch}..origin/{branch}")
            commits_ahead = repo.iter_commits(f"origin/{branch}..{branch}")
            behind = sum(1 for c in commits_behind)
            ahead = sum(1 for c in commits_ahead)
            if(behind > 0):
                status += f"O-{behind} "
            if(ahead > 0):
                status += f"O+{ahead} "


        #- Check files
        diff = repo.index.diff(None)
        N = sum(1 for c in diff)
        if(N > 0):
            status += f"M+{N} "

        untracked =repo.untracked_files
        UN = sum(1 for c in untracked)
        if(UN > 0):
            status += f"U+{UN} "


        self.comment(f"%-25s%-15s%-15s" % (self.name,self.revision,status))


        if(self.verbose):
            self.indent +=1
            for d in untracked:
                self.warning(self.name + os.path.sep + d + " ?" )
            for d in diff:
                self.warning(self.name + os.path.sep + d.a_path + " "  + d.change_type)
            self.indent -=1

        #if(not isClean):
        #    print(status)

    def update(self):

        if(not self.exists()):
            self.warning(f"%-25s%-15s%-15s" % (self.name,"","uncloned"))
            self.clone()
            return

        self.warning(f"%-25s%-15s%-15s" % (self.name,"","updating"))
        repo = git.Repo(self.name)
        repo.git.fetch()
        repo.git.checkout(self.revision)
        try:
            repo.git.pull()
            
        except Exception as e:
            print(self.name + " is detached at " + self.revision + ", no need to pull")


class Config(cicconf.Command):
    def __init__(self,filename,verbose=False):
        super().__init__(verbose)
        self.filename = filename
        self.children = dict()
        self.options = dict()

    def read(self):

        if(not os.path.exists(self.filename)):
            self.error(f"Could not find {self.filename}")
            return False

        with open(self.filename) as fi:
            self.config = yaml.safe_load(fi)

        for k in self.config:
            if(k == "options"):
                self.options = self.config[k]
            else:
                self.children[k] = Repo(k,self.config[k],self.verbose)

        return True

    def clone(self,useHttps,doOnClone=True):
        for name,c in self.children.items():
            c.clone(useHttp=useHttps,doOnClone=doOnClone)

    def newIp(self,name):
        #- Check for errors
        error = False
        if("project" not in self.options):
            self.error(f"project option not defined in {self.filename}")
            error = True
        if("technology" not in self.options):
            self.error(f"technology option not defined in {self.filename}")
            error = True

        if("template" not in self.options):
            self.error(f"template option not defined in {self.filename}")
            error = True
        if(not error and "ip" not in self.options["template"]["ip"]):
            self.error(f"no 'ip' option defined for options->template in {self.filename}")
            error = True
        if(error):
            return

        #- Run template
        ip = self.options["project"] + "_" + name + "_" + self.options["technology"]
        iptemplate = self.options["template"]["ip"]
        cmd = cicconf.CmdIp(ip.upper(),iptemplate)
        cmd.run()

    def status(self):
        self.comment("O = Commits behind/ahead branch","yellow")
        self.comment("M = Modified files","yellow")
        self.comment("U = Untracked files","yellow")
        self.comment("%-25s%-15s%-15s" %("Name","Revision","Status"))
        self.comment("-"*55)
        for name,c in self.children.items():
            c.status()

    def update(self):

        for name,c in self.children.items():
            c.update()



import setuptools

# The setup.py file was used from old versions of pip
# After PEP 518 (don't know which version)
# it was not used anymore.
# Basically, the pyproject.toml and setup.py must be kept in sync


with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cicconf",
    version="0.1.6",
    author="Carsten Wulff",
    author_email="carsten@wulff.no",
    description="Configuration and Revision tool ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wulffern/cicconf",
    packages=setuptools.find_packages(),
    python_requires='>=3.8',
    entry_points = {'console_scripts': [
        'cicconf = cicconf.entry:cli',
    ]},
    install_requires = 'matplotlib click GitPython'.split(),
    classifiers = [
        'Development Status :: 1 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        "License :: OSI Approved :: MIT License",
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering',
    ],
)
